/*
# Copyright (c) 2003-2004 by the University of Southern California.
# All rights reserved.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation in source and binary forms for non-commercial purposes and
# without fee is hereby granted, provided that the above copyright notice
# appear in all copies and that both the copyright notice and this permission
# notice appear in supporting documentation, and that any documentation,
# advertising materials, and other materials related to such distribution and
# use acknowledge that the software was developed by the University of
# Southern California, Information Sciences Institute.  The name of the
# University may not be used to endorse or promote products derived from this
# software without specific prior written permission.
#
# THE UNIVERSITY OF SOUTHERN CALIFORNIA MAKES NO REPRESENTATIONS ABOUT THE
# SUITABILITY OF THIS SOFTWARE FOR ANY PURPOSE.  THIS SOFTWARE IS PROVIDED "AS
# IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT
# LIMITATION, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE.
#
# Other copyrights might apply to parts of this software and are so noted when
# applicable.
# ----------------------------------------------------------------------------
#
# Primary author: Purushotham Kamath
#                 July 2003
#This material is based upon work supported by the Defense Advanced Research 
#Projects Agency under contract no. N66001-02-1-8939 issued by the Space and 
#Naval Warfare Systems Center (SPAWAR).  Any opinions, findings, and conclusions 
#or recommendations expressed in this material are those of the author(s) and 
#do not necessarily reflect the views of the Defense Advanced Research Projects
#Agency, SPAWAR, or the U.S. Government.
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <packet.h>
#include <eventhandler.h>
#include <queue.h>
#include <scheduler.h>
#include <ipport.h>
#include <sense.h>
#include <line.h>
#include <debug.h>

extern Line *l;
extern Scheduler * sched;
extern int alg;
extern IPPort *ip;

Sense::Sense(double b, int s, int w, int o, int m, double bac): bw(b), spreading(s), window(w), drops(0), othreshold(o), mthreshold(m)
{

	totalpktssent = 0;
	totaldefers = 0;
	totalsenses = 0;

//	backoff = 0.5* 1000 * 8/ (bw /(double)spreading ) ;  //Back off = approx one packet length
	backoff = bac;

	double delay = spreading/bw;
	sched->scheduleevent(this, NULL, delay);

	linestate =0;
	newlinestate =0;
	sensecount =0;

	thstate = new int[30];
	newthstate = new int[30];
	for(int count =0; count < 30; count++)
	{
		thstate[count] =0;
		newthstate[count] =0;
	}
	codecount = new int[ip->numcodes *spreading];
	newcodecount = new int[ip->numcodes *spreading];
	for(int count =0; count < ip->numcodes; count++)
	{
		codecount[count] = 0;
		newcodecount[count] = 0;
	}
}

Sense::~Sense()
{
}

//The Sense message handling function: It queues the packet, checks if the port is busy, processes it if not.
int Sense::process(Packet  *packet)
{

	dprintf("sense.cc:Sense::process(): Starting sensing: Packet id: %d\n", packet->id);  
	mac(packet);

}



//The Sense event handling fucntion: Dequeues a packet, processes it, 
//checks if an input interface is waiting and schedules it if it is.
void Sense::handleevent(Packet *packet)
{

	if(packet != NULL)
	{

		switch(packet->state)
		{

			case WAITINGFORTX:
			{
				dprintf("Sense::handleevent(): Sensing complete, transmitting packet id: %d\n", packet->id);

				totalpktssent++;
				totaldefers+=packet->defers;
				
				l->process(packet);	
				break;
			}
			case BACKOFF:
			{
				process(packet);
				break;
			}
		}
	}	
//Sensing
	if(packet == NULL)
	{

	if(sensecount >= window)
	{
		linestate = newlinestate;
		newlinestate = 0;

		for(int count =0; count < ip->numcodes; count++)
			for(int count2 =0 ;count2 < spreading; count2++)
			{
				codecount[count *spreading + count2] = newcodecount[count *spreading + count2];
				newcodecount[count *spreading + count2] = 0;
			}
		for(int count =(int)spreading -1 ; count >= 0; count--)
		{
			thstate[count] =newthstate[count];
			newthstate[count] = 0;
		}



		sensecount=0;
	}
		
	sense();
	dprintf("Sense::handleevent(): Sensing in progress for packet id: sense count: %d \n");
		double delay = (double)spreading/bw;
		nextsensingtime = sched->currenttime + delay;

//Continue sensing only if there are packets to be sent - else the simualtion 
//will run ad infinitum
		sched->scheduleevent(this, NULL, delay);

	}
}



int Sense::mac(Packet *p)
{

//Aloha CDMA: Set window = 0
	if (alg == ALOHA)
	{

	double prop = 1000 / 2e8;
		p->state = WAITINGFORTX;
		//printf("Sense::handleevent(): Setting txtime:: %0.9f\n", p->txtime);
		p->txtime = sched->currenttime;

//Bug fix: Prop Delay should be in line element
		double delay = 0;
		sched->scheduleevent(this, p, delay);
		return 0;
	}


	else if(alg == SELFISH)
	{

//Mac algorithm
	dprintf("Sense::handleevent(): Selfish alg Sensing complete, media access packet id: %d\n", p->id);

	int count =0;


	int codeword = p->codeword;

	dprintf("Sense::handleevent(): Linestate: %d\n", linestate);
	while((linestate & codeword) == codeword)
	{
		dprintf("Sense::handleevent(): Codeword: %d\n", codeword);
		dprintf("Sense::handleevent(): Linestate & Codeword: %d\n", linestate & codeword);
		codeword = rightshift(codeword, 1);
		count++;
		if(count > spreading)
			break;
	}

	if(count > spreading)
	{
		printf("Sense::handleevent(): Deferring\n");
		p->defers++;
		if(p->defers > 10)
		{
			drops++;
			delete p;
			return 0;
		}

		p->state = BACKOFF;
		sched->scheduleevent(this, p, backoff);
	//	process(p);
		return 0;

	}

	p->state = WAITINGFORTX;
	double prop = 1000 / 2e8;

	if((sched->currenttime -nextsensingtime) > 1e-9)
	{
		printf("Error: Schedulin sensing times\n");
		exit(-1);
	}


	int departure = count;
	dprintf("Deaprture: %d\n", count);
//Bug fix: Prop Delay should be in line element
	p->txtime = nextsensingtime + (double)departure / bw;
	dprintf("Next sesnign time%.12f\n", nextsensingtime);
	dprintf("Prop %.12f\n", prop);
	dprintf("Deaprt/bw: %.12f\n", (double)departure/bw);

	dprintf("Sense::handleevent(): Setting txtime:: %0.9f\n", p->txtime);
	//printf("sENse::handleevent(): Departure instant: %d\n", departure);
	double delay = p->txtime - sched->currenttime;
	sched->scheduleevent(this, p, delay);

	}
	
	else if(alg == THRESHOLD)
	{

//Mac algorithm
	dprintf("Sense::handleevent(): Sensing complete, media access packet id: %d\n", p->id);

	int count =0;


	int codeword = p->codeword;
	dprintf("Codeword: %d\n", p->codeword);		
	int departure = 0;
	int overlapcount =0;	
	for(int count =0 ;count < spreading; count++)
	{
		overlapcount =0;	

		dprintf("Sense::handleevent(): Linestate: %d\n", linestate);
		dprintf("Sense::handleevent(): Th state: :\n");
	
		if ((linestate & codeword) == codeword)
		{
			dprintf("Selfish fail\n");
			codeword = rightshift(codeword, 1);
			continue;
		}

		int mcount = 0;
		for(int count2 =(int)spreading -1 ; count2 >= 0; count2--)
		{
			int chip = codeword & (int)pow(2, count2);
			::dprintf("%d ", thstate[count2]);

			if(chip != 0)
			{

				if((thstate[count2] > mthreshold) || ((thstate[count2] +1) > mthreshold))
				{
					//Magnitude threshold has been exceeded, check next code shift
					mcount = thstate[count2] + 1;
					dprintf("Magnitude count exceeded: %d\n", mcount);
					break;
				}
				if(((thstate[count2] +1) == mthreshold) || ((thstate[count2]) ==  mthreshold))
					overlapcount++;
			}
		}

		dprintf("Overlap count: %d\n ", overlapcount);
		dprintf("Magnitdue count: %d\n ", mcount);

		if((overlapcount < othreshold) && (mcount < mthreshold))
		{
			departure = count;
			goto send;
		}


		dprintf("Overlap  count exceeded: %d\n", overlapcount);
		codeword = rightshift(codeword, 1);
	}

defer:
		dprintf("Sense::handleevent(): Deferring\n");
		p->defers++;
		if(p->defers > 10)
		{
			drops++;
			delete p;
			return 0;
		}

		dprintf("Backing off: %.12f\n", backoff);
		p->state = BACKOFF;
		sched->scheduleevent(this, p, backoff);
	//	process(p);
		return 0;


send:

	dprintf("New state: \n");
	int tmpcode = rightshift(p->codeword, departure);
	for(int count =(int)spreading -1 ; count >= 0; count--)
	{
			if((tmpcode & (int)pow(2, count) ) != 0)
				thstate[count]++;
			::dprintf("%d ", thstate[count]);
	}
	p->state = WAITINGFORTX;

	double prop = 1000 / 2e8;

/*
	if((sched->currenttime -nextsensingtime) > 1e-9)
	{
		printf("Error: Cuurent time: %.12f Sensing time: %.12f\n", sched->currenttime, nextsensingtime);
		printf("Error: Schedulin sensing times\n");
		exit(-1);
	}
*/
   
//Bug fix: Prop Delay should be in line element
	p->txtime = nextsensingtime +  (double)departure / bw; 
	dprintf("Sense::handleevent(): Setting txtime:: %0.9f\n", p->txtime);
	dprintf("Sense::handleevent(): Departure instant: %d\n", departure);
	double delay = p->txtime - sched->currenttime; 
	sched->scheduleevent(this, p, delay);

	}
	else if(alg == CODEEST)
	{

//Mac algorithm
	dprintf("Sense::handleevent(): Sensing complete, media access packet id: %d\n", p->id);

	int limit =10;
	int index[limit];


	int departure = 0;
	int pcode = p->codeword;


	dprintf("Codeword: %d\n", p->codeword);		
	for(int count0 = 0; count0 < limit; count0++)
		index[count0] =-1;


	for(int count =0; count < ip->numcodes; count++)
	{
		for(int count2 =0 ;count2 < spreading; count2++)
		{

				if(codecount[count *spreading +count2] > index[0])
				{
					index[0] = count*spreading+count2;
								
				
					for(int count3 =0; count3 < limit-1; count3++)
					{
						if(index[count3] > index[count3+1])
						{
							int tmp = index[count3+1];
							index[count3+1] = index[count3];
							index[count3] = tmp; 
						}
					}
				}
		}
	}
	for(int count0 = 0; count0 < limit; count0++)
		dprintf("Index[%d]: %d: %d\n", count0, index[count0], codecount[index[count0]]);

	if(codecount[index[limit-1]] > 0)
	{

	for (int count =limit -1 ; count >=0; count--)
	{

		dprintf("Checking\n");
		dprintf("Count: %d Limit %d\n", count, limit);
		if(codecount[index[count]] <= 0)
			continue;	

		int codeindex = int (floor((double)index[count]/(double)spreading));
		int shift = index[count] % spreading;
		dprintf("Possible code: %d Shifted by %d\n", ip->code[codeindex], shift);
		dprintf("Possible code: %d\n", rightshift(ip->code[codeindex], shift));
	
		while  ((linestate & pcode) == pcode)
		{
			dprintf("Selfish fail\n");
			pcode = rightshift(pcode, 1);
			departure++;
			if(departure > spreading)
				goto checkcomplete;
		}

		int line =0;
		for(int count2 = limit- 1; count2 >= 0; count2--)
		{
			if(count == count2)
				continue;
			int codeindex2 = int (floor((double)index[count2]/(double)spreading));
			int shift2 = index[count2] % spreading;
			line = line | rightshift(ip->code[codeindex2], shift2);

		}
		line = line | pcode;
		dprintf("LIne: %d\n", line);
		dprintf("Count: %d Limit %d\n", count, limit);

		if((line & ip->code[codeindex]) == ip->code[codeindex])
		{
			dprintf("Interfeence: Shifting\n");
			pcode = rightshift(pcode, 1);
			departure++;
			if(departure > spreading)
				goto checkcomplete;
		}
	}
	}


checkcomplete:
	if(departure >  spreading)
	{

		dprintf("Sense::handleevent(): Deferring\n");
		p->defers++;
		if(p->defers > 100)
		{
			drops++;
			delete p;
			return 0;
		}

		dprintf("Backing off: %.12f\n", backoff);
		p->state = BACKOFF;
		sched->scheduleevent(this, p, backoff);
	//	process(p);
		return 0;
	}
	else
	{

	p->state = WAITINGFORTX;

	double prop = 1000 / 2e8;

	if((sched->currenttime -nextsensingtime) > 1e-9)
	{
		printf("Error: Schedulin sensing times\n");
		exit(-1);
	}


//Bug fix: Prop Delay should be in line element
	p->txtime = nextsensingtime + (double)departure / bw; 
	dprintf("Sense::handleevent(): Setting txtime:: %0.9f\n", p->txtime);
	dprintf("Sense::handleevent(): Departure instant: %d\n", departure);
	double delay = p->txtime - sched->currenttime;
	sched->scheduleevent(this, p, delay);

	}

	}

}

int Sense::sense()
{

	dprintf("sense.cc:Sense::sense(): Sensing \n" );
	sensecount++;

	Queue *q = l->queue();

	int line = 0;

	int thline[spreading];
	for(int count=0; count < spreading; count++)
		thline[count] = 0;

	Packet *qp = q->head;
	while (qp != NULL)
	{
		dprintf("sense.cc:Sense::sense(): Packet on line: \n");


//0 or 1
		double one = (random()/(pow(2, 31) -1)); 
		if(one < 0.5)
		{
			qp=qp->next;
			continue;
		}	


		double timeshift = sched->currenttime - qp->txtime;	
		dprintf("sense.cc:Sense::sense(): Packet id:  %d\n", qp->id);
		dprintf("sense.cc:Sense::sense(): Tx time: %.9f\n", qp->txtime);
		dprintf("sense.cc:Sense::sense(): Current time: %.9f\n", sched->currenttime);
		double chipshift = timeshift *bw; 

		dprintf("sense.cc:Sense::sense(): Chip shift: %d\n", (int) rint(chipshift));
		dprintf("sense.cc:Sense::sense(): Spreading: %d\n", spreading);
	//	int chiprem = ((int) rint(chipshift)) % spreading;

	double qt = floor(chipshift/(double)(spreading));
	dprintf("qt: %.12f\n", qt); 
    double remainder = chipshift - ((double)spreading)*qt;
	dprintf("remainder: %.12f\n", remainder); 
	int chiprem = ((int) rint(remainder));
	

	dprintf("sense.cc:Sense:sense(): Mod Chip shift: %d\n", chiprem);

		int shifted = leftshift(qp->codeword, chiprem);

		if(alg == THRESHOLD)
		{
		for(int count =(int)spreading -1 ; count >= 0; count--)
		{
			if(((shifted & (int)pow(2, count)) != 0))
				thline[count]++;
			dprintf("%d \n", thline[count]);
		}	
		}	


	line = line | shifted;


	dprintf("sense.cc:Sense:sense(): Line: %d\n", line);


		qp=qp->next;
	}
		


//Sensing algorithm
	newlinestate = newlinestate | line;

	if(alg == THRESHOLD)
	{

		dprintf("sense.cc:Sense:sense(): Thstate\n");
		for(int count =(int)spreading -1 ; count >= 0; count--)
		{
			if( newthstate[count] < thline[count])
				newthstate[count] = thline[count]; 
			dprintf("%d \n", newthstate[count]);
		}
	}

	dprintf("sense.cc:Sense:sense(): Line state: %d\n", linestate);

	if(alg == CODEEST)
	{
		for(int count =0; count < ip->numcodes;count++)
		{
			int testcode = ip->code[count];
			for(int count2 =0; count2 < spreading; count2++)
			{
				testcode = rightshift(testcode, 1);
				if((line & testcode) == testcode)
				{
					dprintf("Identified apossibility: %d\n",  testcode);
					newcodecount[count *spreading  + count2]++;
				}
			}
		}
	}

}



void Sense::dumpstats()
{

	double avgdefer = (double)totaldefers/(double)totalpktssent;
	double avgsense = (double)totalsenses/(double)(totalpktssent);
	//printf("%d %.6f %.6f\n", drops, avgdefer, avgsense);
	printf("Average number of defers: %.6f\n", avgdefer);
	printf("Number of packets dropped due to exceeding the sensing limit: %d\n", drops);

}


int Sense:: rightshift( int code, int shift)
{
    int shifted =code;
	for(int count=0; count< shift; count++)
	{
		int lsb = shifted & 1;
//      printf("lsb: %d\n", lsb);
		if (lsb != 1)
			shifted = shifted >> 1;
		else
			shifted = (shifted >> 1) | (int)(pow(2, spreading - 1)) ;
//        dprintf("Right shifted: %d\n", shifted);
	}
	return shifted;
}

int Sense:: leftshift( int code, int shift)
{
    int shifted =code;
	for(int count=0; count< shift; count++)
	{
		int msb = shifted & (int)pow(2, spreading -1);
//      printf("lsb: %d\n", lsb);
		if (msb ==0)
			shifted = shifted << 1;
		else
			shifted = (shifted << 1) | 1 ;

		shifted = shifted & ((int)pow(2, spreading) - 1);
        dprintf("Left shifted: %d\n", shifted);
	}
	return shifted;
}


int Sense::chipcount( int code)
{
    int chipcount =0;
	int shifted =code;
	for(int count=0; count< spreading; count++)
	{
		int msb = shifted & (int)pow(2, spreading -1);
//      //printf("lsb: %d\n", msb); 
		if (msb !=0)
		chipcount++;
		shifted = shifted << 1;
												
		shifted = shifted & ((int)pow(2, spreading) - 1);
	}
	return chipcount;
}

